import pandas as pd
from os.path import isfile
import fileManagement
import numpy as np


def parseMD(results, manifest = None, out_file = None):
    if fileManagement.check_file(out_file):
        return fileManagement.load_data(out_file)
    
    if not isinstance(results, list):
        raise AssertionError("MD results input must be list")
    
    if len(results) == 0:
        raise AssertionError("'results' contains no detections")
        
    df = pd.DataFrame()
    for dictionary in results:
        detections = dictionary['detections']
        for detection in detections:
            bbox = detection['bbox']
            data = {'file': dictionary['file'], 'max_detection_conf': dictionary['max_detection_conf'],
                    'category': detection['category'], 'conf': detection['conf'], 'bbox1': bbox[0],
                    'bbox2': bbox[1],
                    'bbox3': bbox[2], 'bbox4': bbox[3]}
            df = df.append(data, ignore_index=True)
       
    # adjust boxes with 2% buffer from image edge 
    df.loc[df["bbox1"] > 0.98, "bbox1"] = 0.98
    df.loc[df["bbox2"] > 0.98, "bbox2"] = 0.98
    df.loc[df["bbox3"] > 0.98, "bbox3"] = 0.98
    df.loc[df["bbox4"] > 0.98, "bbox4"] = 0.98
    
    df.loc[df["bbox1"] < 0.02, "bbox1"] = 0.02
    df.loc[df["bbox2"] < 0.02, "bbox2"] = 0.02
    df.loc[df["bbox3"] < 0.02, "bbox3"] = 0.02
    df.loc[df["bbox4"] < 0.02, "bbox4"] = 0.02
    
    
    if isinstance(manifest, pd.DataFrame):
        df = manifest.merge(df, left_on="Frame", right_on="file")
        
    if out_file:
        fileManagement.save_data(df, out_file)
            
    return df
    
# def parseMDjson
def applyPredictions(animals, predictions, class_file, out_file = None, counts = False):
    if fileManagement.check_file(out_file): 
        return fileManagement.load_data(out_file)

    if not isinstance(animals, pd.DataFrame):
        raise AssertionError("'animals' must be DataFrame.")

    if not isfile(class_file):
        raise AssertionError("The given class file does not exist.")

    # Format Classification results
    table = pd.read_table(class_file, sep=" ", index_col=0)

    animals['prediction'] = [table['x'].values[int(np.argmax(x))] for x in predictions]
    animals['confidence'] = [np.max(x) for x in predictions] * animals["conf"]
    
    return animals


"""
Module: animl.plot_boxes
Functionality to draw bounding boxes and labels provided image DataFrame.
"""

import cv2


def draw_bounding_boxes(row, box_number, image_output_path, prediction):
    # pylint: disable=R0914
    """
    Draws bounding boxes and labels on image DataFrame.
    Args:
        image : DataFrame containing image data - coordinates and predictions.
            The DataFrame should have the following columns:
            - 'Frame': Filename or path to the image file.
            - 'bbox1': Normalized x-coordinate of the top-left corner.
            - 'bbox2': Normalized y-coordinate of the top-left corner.
            - 'bbox3': Normalized width of the bounding box (range: 0-1).
            - 'bbox4': Normalized height of the bounding box (range: 0-1).
            - 'prediction': Object prediction label for the bounding box.
        n (int): Number used for generating the output image filename.
        imageOutputPath (str): Output directory to saved images.
    Returns:
        None
    """
    img = cv2.imread(row["Frame"])  # Use row directly without indexing
    height, width, _ = img.shape
    left = int(row['bbox1'] * width)
    top = int(row['bbox2'] * height)
    right = int((row['bbox1'] + row['bbox3']) * width)
    bottom = int((row['bbox2'] + row['bbox4']) * height)
    label = row['prediction']
    text_size, _ = cv2.getTextSize(label, cv2.FONT_HERSHEY_DUPLEX, 1, 1)
    text_size_width, text_size_height = text_size
    thick = int((height + width) // 900)
    box_right = (right if (right - left) < (text_size_width * 3)
                else left + (text_size_width * 3))
    cv2.rectangle(img, (left, top), (right, bottom), (90, 255, 0), thick)
    cv2.rectangle(img, (left, top),
                  (box_right, top - (text_size_height * 2)),
                  (90, 255, 0), -1)
    if prediction:
        cv2.putText(img, label, (left, top - 12), 0, 1e-3 * height,
                    (0, 0, 0), thick // 3)
        
    filename = image_output_path + str(box_number) + ".jpg"
    print(filename)
    cv2.imwrite(filename, img)

    
def demo_boxes(manifest, min_conf = 0.9, prediction = True):
    # pylint: disable=R0914
    """
    Draws bounding boxes and labels on image DataFrame.
    Args:
        image : DataFrame containing image data - coordinates and predictions.
            The DataFrame should have the following columns:
            - 'Frame': Filename or path to the image file.
            - 'bbox1': Normalized x-coordinate of the top-left corner.
            - 'bbox2': Normalized y-coordinate of the top-left corner.
            - 'bbox3': Normalized width of the bounding box (range: 0-1).
            - 'bbox4': Normalized height of the bounding box (range: 0-1).
            - 'prediction': Object prediction label for the bounding box.
        n (int): Number used for generating the output image filename.
        imageOutputPath (str): Output directory to saved images.
    Returns:
        None
    """
    images = manifest["FilePath"].unique()
    
    for image_path in images:
        #display the image, wait for key
        img = cv2.imread(image_path)  # Use row directly without indexing
        cv2.namedWindow("Display", cv2.WINDOW_NORMAL)
        cv2.imshow('Display', img)
        cv2.waitKey(0)
        
        boxes = manifest[manifest["FilePath"] == image_path]
        print(boxes)
        for _,row in boxes.iterrows():
            confidence = row["confidence"]
            if confidence >= min_conf:
                height, width, _ = img.shape
                left = int(row['bbox1'] * width)
                top = int(row['bbox2'] * height)
                right = int((row['bbox1'] + row['bbox3']) * width)
                bottom = int((row['bbox2'] + row['bbox4']) * height)
                label = row['prediction']
                text_size, _ = cv2.getTextSize(label, cv2.FONT_HERSHEY_DUPLEX, 1, 1)
                text_size_width, text_size_height = text_size
                thick = int((height + width) // 900)
                box_right = (right if (right - left) < (text_size_width * 3)
                            else left + (text_size_width * 3))

                cv2.rectangle(img, (left, top), (right, bottom), (90, 255, 0), thick)

                if prediction:

                    cv2.rectangle(img, (left, top),
                              (box_right, top - (text_size_height * 3)),
                              (90, 255, 0), -1)
                    cv2.putText(img, label, (left, top - 12), 0, 1e-3 * height,
                                (0, 0, 0), thick // 3)
                cv2.imshow('Display', img)
                cv2.waitKey(0)
            
            else:
                continue

    cv2.destroyAllWindows()


import os
import cv2
from random import randrange
import multiprocessing as mp
import fileManagement
import pandas as pd
from numpy import vstack

# TODO 
#
# 1. check if video file is corrupt
# 2. implement checkpoint


# Extract frames from video for classification
#
# @param file_path dataframe of videos
# @param out_dir directory to save frames to
# @param fps frames per second, otherwise determine mathematically
# @param frames number of frames to sample
#
# @return dataframe of still frames for each video
def extractImages(file_path, out_dir, fps=None, frames=None):
    cap = cv2.VideoCapture(file_path)
    filename = os.path.basename(file_path)
    filename, extension = os.path.splitext(filename)
    uniqueid = '{:05}'.format(randrange(1, 10 ** 5))
    frames_saved = []

    if fps is None:  # select set number of frames
        frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        frame_capture = 0
        increment = int(frame_count / frames)
        while cap.isOpened() and (len(frames_saved) < frames):
            cap.set(cv2.CAP_PROP_POS_FRAMES, frame_capture)
            ret, frame = cap.read()
            if not ret:
                break

            out_path = (out_dir + filename + "-" +
                        uniqueid + "-" +
                        str(frame_capture) + '.jpg')
            cv2.imwrite(out_path, frame)
            frames_saved.append([out_path,file_path])
            frame_capture += increment
            
            
    else:  # select by fps
        frame_capture = 0
        while cap.isOpened():
            ret, frame = cap.read()
            cap.set(cv2.CAP_PROP_POS_MSEC, (frame_capture * 1000))
            if not ret:
                break

            out_path = (out_dir + filename + "-" +
                        '{:05}'.format(randrange(1, 10 ** 5)) + "-" +
                        str(frame_capture) + '.jpg')
            cv2.imwrite(out_path, frame)
            frames_saved.append([out_path,file_path])
            frame_capture += fps

    cap.release()
    cv2.destroyAllWindows()
    
    return frames_saved


# Extract frames from video for classification
#
# @param files dataframe of videos
# @param out_dir directory to save frames to
# @param outfile file to which results will be saved
# @param format output format for frames, defaults to jpg
# @param fps frames per second, otherwise determine mathematically
# @param frames number of frames to sample
# @param parallel Toggle for parallel processing, defaults to FALSE
# @param workers number of processors to use if parallel, defaults to 1
# @param checkpoint if not parallel, checkpoint ever n files, defaults to 1000
#
# @return dataframe of still frames for each video
def images_from_videos(files, out_dir, out_file=None, format="jpg",
                       fps=None, frames=None, parallel=False, 
                       workers=mp.cpu_count(), checkpoint=1000):
    if fileManagement.check_file(out_file):
        return fileManagement.load_data(out_file)
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)
    if (fps is not None) and (frames is not None):
        print("If both fps and frames are defined fps will be used.")
    if (fps is None) and (frames is None):
        raise AssertionError("Either fps or frames need to be defined.")
    # if fileManagement.check_file(outfile): 
    #    temporary = fileManagement.load_data(outfile)
    #    check against checkpoint   
        
    images = files[files["FileName"].apply(
        lambda x: os.path.splitext(x)[1].lower()).isin([".jpg",".jpeg",".png"])]
    images = images.assign(Frame = images["FilePath"])
    
    videos =  files[files["FileName"].apply(
        lambda x: os.path.splitext(x)[1].lower()).isin([".mp4", ".avi", ".mov", ".wmv",
                                                        ".mpg", ".mpeg", ".asf", ".m4v"])]
    if not videos.empty:
        if parallel:
            pool = mp.Pool(workers)

            video_frames = vstack([pool.apply(extractImages, 
                                               args=(video, out_dir, fps, frames)) for 
                                               video in videos["FilePath"]])

            video_frames = pd.DataFrame(video_frames, columns = ["Frame","FilePath"])

            pool.close()

        else:
            video_frames = []
            for i, video in videos.iterrows():
                video_frames += extractImages(video["FilePath"],out_dir=out_dir, 
                                         fps=fps, frames=frames)

                if (i % checkpoint == 0) and (outfile is not None):
                    fileManagement.save_data(images,outfile)

            video_frames = pd.DataFrame(video_frames, columns = ["Frame","FilePath"])

        videos = videos.merge(video_frames, on="FilePath")

    allframes = pd.concat([images,videos])

    if (out_file is not None):
        fileManagement.save_data(allframes,out_file) 

    return allframes


from tensorflow.keras.utils import Sequence
import numpy as np
from PIL import Image, ImageOps, ImageFile
import tensorflow as tf

ImageFile.LOAD_TRUNCATED_IMAGES = True


def resize_with_padding(img, expected_size):
    if img.size[0] == 0 or img.size[1] == 0:
        return img
    if img.size[0] > img.size[1]:
        new_size = (expected_size[0],
                    int(expected_size[1] * img.size[1] / img.size[0]))
    else:
        new_size = (int(expected_size[0] * img.size[0] / img.size[1]),
                    expected_size[1])
    img = img.resize(new_size, Image.BILINEAR)  # NEAREST BILINEAR
    delta_width = expected_size[0] - img.size[0]
    delta_height = expected_size[1] - img.size[1]
    pad_width = delta_width // 2
    pad_height = delta_height // 2
    padding = (pad_width, pad_height,
               delta_width - pad_width, delta_height - pad_height)
    return ImageOps.expand(img, padding)


class GenerateCropsFromFile(Sequence):
    def __init__(self, x, filecol = 'file', resize=299, buffer=0, batch=32, standardize=True):
        self.x = x
        self.filecol = filecol
        self.resize = int(resize)
        self.buffer = buffer
        self.batch = int(batch)
        self.standardize = standardize

    def __len__(self):
        return int(np.ceil(len(self.x.index) / float(self.batch)))

    def __getitem__(self, idx):
        imgarray = []
        for i in range(min(len(self.x.index), idx * self.batch),
                       min(len(self.x.index), (idx + 1) * self.batch)):
            try:
                file = self.x[self.filecol].iloc[i]
                img = Image.open(file)
            except OSError:
                continue
            width, height = img.size

            bbox1 = self.x['bbox1'].iloc[i]
            bbox2 = self.x['bbox2'].iloc[i]
            bbox3 = self.x['bbox3'].iloc[i]
            bbox4 = self.x['bbox4'].iloc[i]

            left = width * bbox1
            top = height * bbox2
            right = width * (bbox1 + bbox3)
            bottom = height * (bbox2 + bbox4)

            left = max(0, left - self.buffer)
            top = max(0, top - self.buffer)
            right = min(width, right + self.buffer)
            bottom = min(height, bottom + self.buffer)

            img = img.crop((left, top, right, bottom))
            img = img.resize((self.resize, self.resize))
            img = tf.keras.utils.img_to_array(img)
            imgarray.append(img)

        return np.asarray(imgarray)


import os
import glob
from PIL import Image

DEFAULT_COLORS = [
    'AliceBlue', 'Red', 'RoyalBlue', 'Gold', 'Chartreuse', 'Aqua', 'Azure',
    'Beige', 'Bisque', 'BlanchedAlmond', 'BlueViolet', 'BurlyWood', 'CadetBlue',
    'AntiqueWhite', 'Chocolate', 'Coral', 'CornflowerBlue', 'Cornsilk', 'Crimson',
    'Cyan', 'DarkCyan', 'DarkGoldenRod', 'DarkGrey', 'DarkKhaki', 'DarkOrange',
    'DarkOrchid', 'DarkSalmon', 'DarkSeaGreen', 'DarkTurquoise', 'DarkViolet',
    'DeepPink', 'DeepSkyBlue', 'DodgerBlue', 'FireBrick', 'FloralWhite',
    'ForestGreen', 'Fuchsia', 'Gainsboro', 'GhostWhite', 'GoldenRod',
    'Salmon', 'Tan', 'HoneyDew', 'HotPink', 'IndianRed', 'Ivory', 'Khaki',
    'Lavender', 'LavenderBlush', 'LawnGreen', 'LemonChiffon', 'LightBlue',
    'LightCoral', 'LightCyan', 'LightGoldenRodYellow', 'LightGray', 'LightGrey',
    'LightGreen', 'LightPink', 'LightSalmon', 'LightSeaGreen', 'LightSkyBlue',
    'LightSlateGray', 'LightSlateGrey', 'LightSteelBlue', 'LightYellow', 'Lime',
    'LimeGreen', 'Linen', 'Magenta', 'MediumAquaMarine', 'MediumOrchid',
    'MediumPurple', 'MediumSeaGreen', 'MediumSlateBlue', 'MediumSpringGreen',
    'MediumTurquoise', 'MediumVioletRed', 'MintCream', 'MistyRose', 'Moccasin',
    'NavajoWhite', 'OldLace', 'Olive', 'OliveDrab', 'Orange', 'OrangeRed',
    'Orchid', 'PaleGoldenRod', 'PaleGreen', 'PaleTurquoise', 'PaleVioletRed',
    'PapayaWhip', 'PeachPuff', 'Peru', 'Pink', 'Plum', 'PowderBlue', 'Purple',
    'RosyBrown', 'Aquamarine', 'SaddleBrown', 'Green', 'SandyBrown',
    'SeaGreen', 'SeaShell', 'Sienna', 'Silver', 'SkyBlue', 'SlateBlue',
    'SlateGray', 'SlateGrey', 'Snow', 'SpringGreen', 'SteelBlue', 'GreenYellow',
    'Teal', 'Thistle', 'Tomato', 'Turquoise', 'Violet', 'Wheat', 'White',
    'WhiteSmoke', 'Yellow', 'YellowGreen'
]


def load_image(input_file):
    """
    from CameraTraps/visualization/visualization_utils.py
    """
    image = Image.open(input_file)
    if image.mode not in ('RGBA', 'RGB', 'L'):
        raise AttributeError('Input image {} uses unsupported mode {}'.format(input_file, image.mode))
    if image.mode == 'RGBA' or image.mode == 'L':
        # PIL.Image.convert() returns a converted copy of this image
        image = image.convert(mode='RGB')
    image.load()
    return image


def is_image(s):
    """
    from CameraTraps/MegaDetector
    Check a file's extension against a hard-coded set of image file extensions    '
    """
    image_extensions = ['.jpg', '.jpeg', '.gif', '.png', '.mp4']
    ext = os.path.splitext(s)[1].lower()
    return ext in image_extensions


def crop_image(detections, image, confidence_threshold=0.15, expansion=0):
    """
    Crops detections above *confidence_threshold* from the PIL image *image*,
    returning a list of PIL images.

    *detections* should be a list of dictionaries with keys 'conf' and 'bbox';
    see bbox format description below.  Normalized, [x,y,w,h], upper-left-origin.

    *expansion* specifies a number of pixels to include on each side of the box.
    """

    ret_images = []

    for detection in detections:

        score = float(detection['conf'])

        if score >= confidence_threshold:

            x1, y1, w_box, h_box = detection['bbox']
            ymin,xmin,ymax,xmax = y1, x1, y1 + h_box, x1 + w_box

            # Convert to pixels so we can use the PIL crop() function
            im_width, im_height = image.size
            (left, right, top, bottom) = (xmin * im_width, xmax * im_width,
                                          ymin * im_height, ymax * im_height)

            if expansion > 0:
                left -= expansion
                right += expansion
                top -= expansion
                bottom += expansion

            # PIL's crop() does surprising things if you provide values outside of
            # the image, clip inputs
            left = max(left,0); right = max(right,0)
            top = max(top,0); bottom = max(bottom,0)

            left = min(left,im_width-1); right = min(right,im_width-1)
            top = min(top,im_height-1); bottom = min(bottom,im_height-1)

            ret_images.append(image.crop((left, top, right, bottom)))

        # ...if this detection is above threshold

    # ...for each detection

    return ret_images

    
def render_detection_bounding_boxes(detections, image,
                                    label_map={}, 
                                    classification_label_map=None, 
                                    confidence_threshold=0.15, thickness=4, expansion=0,
                                    classification_confidence_threshold=0.3,
                                    max_classifications=3,
                                    colormap=DEFAULT_COLORS,
                                    textalign=0):
    """
    Renders bounding boxes, label, and confidence on an image if confidence is above the threshold.

    Boxes are in the format that's output from the batch processing API.

    Renders classification labels if present.

    Args:

        detections: detections on the image 
        image: PIL.Image object

        label_map: optional, mapping the numerical label to a string name. The type of the numerical label
            (default string) needs to be consistent with the keys in label_map; no casting is carried out.
            If this is None, no labels are shown.

        classification_label_map: optional, mapping of the string class labels to the actual class names.
            The type of the numerical label (default string) needs to be consistent with the keys in
            label_map; no casting is carried out.  If this is None, no classification labels are shown.

        confidence_threshold: optional, threshold above which the bounding box is rendered.
        
        thickness: line thickness in pixels. Default value is 4.
        
        expansion: number of pixels to expand bounding boxes on each side.  Default is 0.
        
        classification_confidence_threshold: confidence above which classification result is retained.
        
        max_classifications: maximum number of classification results retained for one image.

    image is modified in place.
    """

    display_boxes = []
    
    # list of lists, one list of strings for each bounding box (to accommodate multiple labels)
    display_strs = []  
    
    # for color selection
    classes = []  

    for detection in detections:

        score = detection['conf']
        
        # Always render objects with a confidence of "None", this is typically used
        # for ground truth data.        
        if score is None or score >= confidence_threshold:
            
            x1, y1, w_box, h_box = detection['bbox']
            display_boxes.append([y1, x1, y1 + h_box, x1 + w_box])
            clss = detection['category']
            
            # {} is the default, which means "show labels with no mapping", so don't use "if label_map" here
            # if label_map:
            if label_map is not None:
                label = label_map[clss] if clss in label_map else clss
                if score is not None:
                    displayed_label = ['{}: {}%'.format(label, round(100 * score))]
                else:
                    displayed_label = ['{}'.format(label)]
            else:
                displayed_label = ''

            if 'classifications' in detection:

                # To avoid duplicate colors with detection-only visualization, offset
                # the classification class index by the number of detection classes
                clss = int(detection['classifications'][0][0]) + 3
                classifications = detection['classifications']
                if len(classifications) > max_classifications:
                    classifications = classifications[0:max_classifications]
                    
                for classification in classifications:
                    
                    classification_conf = classification[1]
                    if classification_conf is not None and \
                        classification_conf < classification_confidence_threshold:
                        continue
                    class_key = classification[0]
                    if (classification_label_map is not None) and (class_key in classification_label_map):
                        class_name = classification_label_map[class_key]
                    else:
                        class_name = class_key
                    if classification_conf is not None:
                        displayed_label += ['{}: {:5.1%}'.format(class_name.lower(), classification_conf)]
                    else:
                        displayed_label += ['{}'.format(class_name.lower())]
                    
                # ...for each classification

            # ...if we have classification results
                        
            display_strs.append(displayed_label)
            classes.append(clss)

        # ...if the confidence of this detection is above threshold

    # ...for each detection
    
    display_boxes = np.array(display_boxes)

    draw_bounding_boxes_on_image(image, display_boxes, classes,
                                 display_strs=display_strs, thickness=thickness, 
                                 expansion=expansion, colormap=colormap, textalign=textalign)


# modified from CameraTraps/MegaDetector
from tqdm import tqdm
import time
import json
import humanfriendly
import os
import pandas as pd

from detection import run_detector
from detection import run_detector_batch

def chunks_by_number_of_chunks(ls, n):
    for i in range(0, n):
        yield ls[i::n]

        
def detect_MD_batch(model_file, image_file_names, checkpoint_path=None,
                                confidence_threshold=run_detector.DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD,
                                checkpoint_frequency=-1, results=None, n_cores=1, quiet=False, image_size=None):
    """
    Args
    - model_file: str, path to .pb model file
    - image_file_names: list of strings (image filenames), a single image filename, 
                        a folder to recursively search for images in, or a .json file containing
                        a list of images.
    - checkpoint_path: str, path to JSON checkpoint file
    - confidence_threshold: float, only detections above this threshold are returned
    - checkpoint_frequency: int, write results to JSON checkpoint file every N images
    - results: list of dict, existing results loaded from checkpoint
    - n_cores: int, # of CPU cores to use

    Returns
    - results: list of dict, each dict represents detections on one image
    """
    
    if n_cores is None:
        n_cores = 1
    
    if confidence_threshold is None:
        confidence_threshold=run_detector.DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD
        
    if checkpoint_frequency is None:
        checkpoint_frequency = -1
        
    # Handle the case where image_file_names is not yet actually a list
    if isinstance(image_file_names,str):
        
        # Find the images to score; images can be a directory, may need to recurse
        if os.path.isdir(image_file_names):
            image_dir = image_file_names
            image_file_names = run_detector.ImagePathUtils.find_images(image_dir, True)
            print('{} image files found in folder {}'.format(len(image_file_names),image_dir))
            
        # A json list of image paths
        elif os.path.isfile(image_file_names) and image_file_names.endswith('.json'):
            list_file = image_file_names
            with open(list_file) as f:
                image_file_names = json.load(f)
            print('Loaded {} image filenames from list file {}'.format(len(image_file_names),list_file))
            
        elif isinstance(image_file_names,pd.Series):
            pass
            
        # A single image file
        elif os.path.isfile(image_file_names) and run_detector.ImagePathUtils.is_image_file(image_file_names):
            image_file_names = [image_file_names]
            #print('Processing image {}'.format(image_file_names[0]))
            
        else:        
            raise ValueError('image_file_names is a string, but is not a directory, a json ' + \
                             'list (.json), or an image file (png/jpg/jpeg/gif)')
    
    if results is None:
        results = []

    already_processed = set([i['Frame'] for i in results])

    print('GPU available: {}'.format(run_detector.is_gpu_available(model_file)))
    
    if n_cores > 1 and run_detector.is_gpu_available(model_file):
        print('Warning: multiple cores requested, but a GPU is available; parallelization across ' + \
              'GPUs is not currently supported, defaulting to one GPU')
        n_cores = 1

        
    elif n_cores <= 1:

        # Load the detector
        start_time = time.time()
        detector = run_detector.load_detector(model_file)
        elapsed = time.time() - start_time
        print('Loaded model in {}'.format(humanfriendly.format_timespan(elapsed)))

        # Does not count those already processed
        count = 0

        for im_file in tqdm(image_file_names):

            # Will not add additional entries not in the starter checkpoint
            if im_file in already_processed:
                if not quiet:
                    print('Bypassing image {}'.format(im_file))
                continue

            count += 1

            result = run_detector_batch.process_image(im_file, detector, 
                                   confidence_threshold, quiet=quiet, 
                                   image_size=image_size)
            results.append(result)

            # Write a checkpoint if necessary
            if checkpoint_frequency != -1 and count % checkpoint_frequency == 0:
                
                print('Writing a new checkpoint after having processed {} images since '
                      'last restart'.format(count))
                
                assert checkpoint_path is not None                
                
                # Back up any previous checkpoints, to protect against crashes while we're writing
                # the checkpoint file.
                checkpoint_tmp_path = None
                if os.path.isfile(checkpoint_path):
                    checkpoint_tmp_path = checkpoint_path + '_tmp'
                    shutil.copyfile(checkpoint_path,checkpoint_tmp_path)
                    
                # Write the new checkpoint
                with open(checkpoint_path, 'w') as f:
                    json.dump({'images': results}, f, indent=1)
                    
                # Remove the backup checkpoint if it exists
                if checkpoint_tmp_path is not None:
                    os.remove(checkpoint_tmp_path)
                    
            # ...if it's time to make a checkpoint
            
    else:
        
        # When using multiprocessing, let the workers load the model
        detector = model_file

        print('Creating pool with {} cores'.format(n_cores))

        if len(already_processed) > 0:
            print('Warning: when using multiprocessing, all images are reprocessed')

        pool = workerpool(n_cores)

        image_batches = list(chunks_by_number_of_chunks(image_file_names, n_cores))
        results = pool.map(partial(run_detector_batch.process_images, detector=detector,
                                   confidence_threshold=confidence_threshold,image_size=image_size), 
                           image_batches)

        results = list(itertools.chain.from_iterable(results))

    return results



import argparse
import sys
import json
import os

from datetime import datetime
from animl.ImageCropGenerator import GenerateCropsFromFile
from animl.FileManagement import imagesFromVideos, parseMD, filterImages, symlinkClassification, parseCM
from animl.DetectMD import load_and_run_detector_batch
from tensorflow import keras


def main():
    parser = argparse.ArgumentParser(description='Workflow for running animl')

    parser.add_argument(
        'image_dir',
        type=str,
        help='A directory of images and/or videos to be processed')

    parser.add_argument(
        'megadetector_model',
        type=str,
        help='Path to the megadetector model')

    parser.add_argument(
        'classification_model',
        type=str,
        help='Path to the classification model')

    parser.add_argument(
        'classes',
        type=str,
        help='Path to the classes text file for the classification model')

    parser.add_argument(
        '-output_dir',
        default=None,
        type=str,
        help='Output directory including site name and date')

    parser.add_argument(
        '--fps',
        type=int,
        default=None,
        help='Number of frames per second to be used when extracting images from videos')

    parser.add_argument(
        '--frames',
        type=int,
        default=5,
        help='Number of frames to be extracted from videos')

    parser.add_argument(
        '--parallel',
        action='store_true',
        help='Run the program in a parallel fashion using multiple cores')

    parser.add_argument(
        '--threshold',
        type=float,
        default=0.9,
        help="Confidence threshold between 0 and 1.0. Default is 0.9")

    parser.add_argument(
        '--checkpoint_frequency',
        type=int,
        default=2500,  # approximately every hour
        help='Write results to a temporary file every N images; a value of -1 disables this feature')

    parser.add_argument(
        '--resume_from_checkpoint',
        help='Initiate from the specified checkpoint, which is in the same directory as the output_file specified')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    MegaDetector_file = args.megadetector_model
    Classification_file = args.classification_model
    classes = args.classes
    # load the checkpoint if available
    # relative file names are only output at the end; all file paths in the checkpoint are still full paths
    if args.resume_from_checkpoint:
        assert os.path.exists(args.resume_from_checkpoint), 'File at resume_from_checkpoint specified does not exist'
        with open(args.resume_from_checkpoint) as f:
            saved = json.load(f)
        assert 'images' in saved, \
            'The file saved as checkpoint does not have the correct fields; cannot be restored'
        results = saved['images']
        print('Restored {} entries from the checkpoint'.format(len(results)))
    else:
        results = []
    images = imagesFromVideos(args.image_dir, out_dir=args.output_dir,
                              fps=args.fps, frames=args.frames,
                              parallel=args.parallel)

    # test that we can write to the output_file's dir if checkpointing requested
    if args.checkpoint_frequency != -1:
        checkpoint_path = os.path.join(args.output_dir,
                                       'checkpoint_{}.json'.format(datetime.utcnow().strftime("%Y%m%d%H%M%S")))
        with open(checkpoint_path, 'w') as f:
            json.dump({'images': []}, f)
        print('The checkpoint file will be written to {}'.format(checkpoint_path))
    else:
        checkpoint_path = None

    # run MegaDetector and format results
    detections = detectObjectBatch(images, MegaDetector_file, checkpoint_path, 
                                   args.threshold, args.checkpoint_frequency,
                                   results)

    df = parseMD(detections)
    # filter out all non animal detections
    animalDataframe = getAnimals(df)
    otherDataframe = getEmpty(df)

    # Create generator for classification model
    generator = GenerateCropsFromFile(animalDataframe)

    # Create and predict model
    model = keras.models.load_model(Classification_file)
    predictions = model.predict(generator)

    # Format Classification results
    maxDataframe = applyPredictions(animalDataframe, otherDataframe, predictions, classes)
    # Symlink
    symlinkSpecies(maxDataframe, args.output_dir, classes)


if __name__ == '__main__':
    main()



r"""
Module to run an animal detection model on images.

The main function in this script also renders the predicted
bounding boxes on images and saves the resulting images (with bounding boxes).

This script is not a good way to process lots of images (tens of thousands,
say). It does not facilitate checkpointing the results so if it crashes you
would have to start from scratch. If you want to run a detector (e.g., ours)
on lots of images, you should check out:

1) run_detector_batch.py (for local execution)

2) https://github.com/ecologize/CameraTraps/tree/master/api/batch_processing
   (for running large jobs on Azure ML)

To run this script, we recommend you set up a conda virtual environment
following instructions in the Installation section on the main README, using
`environment-detector.yml` as the environment file where asked.

This is a good way to test our detector on a handful of images and get
super-satisfying, graphical results.  It's also a good way to see how fast a
detector model will run on a particular machine.

If you would like to *not* use the GPU on the machine, set the environment
variable CUDA_VISIBLE_DEVICES to "-1".

If no output directory is specified, writes detections for c:\foo\bar.jpg to
c:\foo\bar_detections.jpg.

This script will only consider detections with > 0.005 confidence at all times.
The `threshold` you provide is only for rendering the results. If you need to
see lower-confidence detections, you can change
DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD.

Reference:
https://github.com/tensorflow/models/blob/master/research/object_detection/inference/detection_inference.py
"""

#%% Constants, imports, environment

import glob
import os
import statistics
import sys
import time
import warnings

import humanfriendly
from tqdm import tqdm

import imageUtils as image_utils

# ignoring all "PIL cannot read EXIF metainfo for the images" warnings
warnings.filterwarnings('ignore', '(Possibly )?corrupt EXIF data', UserWarning)

# Metadata Warning, tag 256 had too many entries: 42, expected 1
warnings.filterwarnings('ignore', 'Metadata warning', UserWarning)

# Numpy FutureWarnings from tensorflow import
warnings.filterwarnings('ignore', category=FutureWarning)

# Useful hack to force CPU inference
# os.environ['CUDA_VISIBLE_DEVICES'] = '-1'


# An enumeration of failure reasons
FAILURE_INFER = 'Failure inference'
FAILURE_IMAGE_OPEN = 'Failure image access'

# Number of decimal places to round to for confidence and bbox coordinates
CONF_DIGITS = 3
COORD_DIGITS = 4

# Label mapping for MegaDetector
DEFAULT_DETECTOR_LABEL_MAP = {
    '1': 'animal',
    '2': 'person',
    '3': 'vehicle'  # available in megadetector v4+
}

# Should we allow classes that don't look anything like the MegaDetector classes?
USE_MODEL_NATIVE_CLASSES = False

# Each version of the detector is associated with some "typical" values
# that are included in output files, so that downstream applications can 
# use them as defaults.
DETECTOR_METADATA = {
    'v2.0.0':
        {'megadetector_version':'v2.0.0',
         'typical_detection_threshold':0.8,
         'conservative_detection_threshold':0.3},
    'v3.0.0':
        {'megadetector_version':'v3.0.0',
         'typical_detection_threshold':0.8,
         'conservative_detection_threshold':0.3},
    'v4.1.0':
        {'megadetector_version':'v4.1.0',
         'typical_detection_threshold':0.8,
         'conservative_detection_threshold':0.3},
    'v5a.0.0':
        {'megadetector_version':'v5a.0.0',
         'typical_detection_threshold':0.2,
         'conservative_detection_threshold':0.05},
    'v5b.0.0':
        {'megadetector_version':'v5b.0.0',
         'typical_detection_threshold':0.2,
         'conservative_detection_threshold':0.05}
}

DEFAULT_RENDERING_CONFIDENCE_THRESHOLD = DETECTOR_METADATA['v5b.0.0']['typical_detection_threshold']
DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD = 0.005

DEFAULT_BOX_THICKNESS = 4
DEFAULT_BOX_EXPANSION = 0
    
#%% Utility functions

def convert_to_tf_coords(array):
    """From [x1, y1, width, height] to [y1, x1, y2, x2], where x1 is x_min, x2 is x_max

    This is only used to keep the interface of the synchronous API.
    """
    x1 = array[0]
    y1 = array[1]
    width = array[2]
    height = array[3]
    x2 = x1 + width
    y2 = y1 + height
    return [y1, x1, y2, x2]


def get_detector_metadata_from_version_string(detector_version):
    """
    Given a MegaDetector version string (e.g. "v4.1.0"), return the metadata for
    the model.  Used for writing standard defaults to batch output files.
    """
    if detector_version not in DETECTOR_METADATA:
        print('Warning: no metadata for unknown detector version {}'.format(detector_version))
        default_detector_metadata = {
            'megadetector_version':'unknown'
        }
        return default_detector_metadata
    else:
        return DETECTOR_METADATA[detector_version]


def get_detector_version_from_filename(detector_filename):
    """
    Get the version number component of the detector from the model filename.  
    
    *detector_filename* will almost always end with one of the following:
        
    megadetector_v2.pb
    megadetector_v3.pb
    megadetector_v4.1 (not produed by run_detector_batch.py, only found in Azure Batch API output files)
    md_v4.1.0.pb
    md_v5a.0.0.pt
    md_v5b.0.0.pt
    
    ...for which we identify the version number as "v2.0.0", "v3.0.0", "v4.1.0", 
    "v4.1.0", "v5a.0.0", and "v5b.0.0", respectively.
    """
    fn = os.path.basename(detector_filename)
    known_model_versions = {'v2':'v2.0.0',
                            'v3':'v3.0.0',
                            'v4.1':'v4.1.0',
                            'v5a.0.0':'v5a.0.0',
                            'v5b.0.0':'v5b.0.0'}
    matches = []
    for s in known_model_versions.keys():
        if s in fn:
            matches.append(s)
    if len(matches) == 0:
        print('Warning: could not determine MegaDetector version for model file {}'.format(detector_filename))
        return 'unknown'
    elif len(matches) > 1:
        print('Warning: multiple MegaDetector versions for model file {}'.format(detector_filename))
        return 'multiple'
    else:
        return known_model_versions[matches[0]]
    

def get_typical_confidence_threshold_from_results(results):
    """
    Given the .json data loaded from a MD results file, determine a typical confidence
    threshold based on the detector version.
    """
    if 'detector_metadata' in results['info'] and \
        'typical_detection_threshold' in results['info']['detector_metadata']:
        default_threshold = results['info']['detector_metadata']['typical_detection_threshold']
    elif ('detector' not in results['info']) or (results['info']['detector'] is None):
        print('Warning: detector version not available in results file, using MDv5 defaults')
        detector_metadata = get_detector_metadata_from_version_string('v5a.0.0')
        default_threshold = detector_metadata['typical_detection_threshold']
    else:
        print('Warning: detector metadata not available in results file, inferring from MD version')
        detector_filename = results['info']['detector']
        detector_version = get_detector_version_from_filename(detector_filename)
        detector_metadata = get_detector_metadata_from_version_string(detector_version)
        default_threshold = detector_metadata['typical_detection_threshold']

    return default_threshold    

    
def is_gpu_available(model_file):
    """Decide whether a GPU is available, importing PyTorch or TF depending on the extension
    of model_file.  Does not actually load model_file, just uses that to determine how to check 
    for GPU availability."""
    
    if model_file.endswith('.pb'):
        import tensorflow.compat.v1 as tf
        gpu_available = tf.test.is_gpu_available()
        print('TensorFlow version:', tf.__version__)
        print('tf.test.is_gpu_available:', gpu_available)                
        return gpu_available
    elif model_file.endswith('.pt'):
        import torch
        gpu_available = torch.cuda.is_available()
        print('PyTorch reports {} available CUDA devices'.format(torch.cuda.device_count()))
        if not gpu_available:
            try:
                # mps backend only available in torch >= 1.12.0
                if torch.backends.mps.is_built and torch.backends.mps.is_available():
                    gpu_available = True
                    print('PyTorch reports Metal Performance Shaders are available')
            except AttributeError:
                pass
        return gpu_available
    else:
        raise ValueError('Unrecognized model file extension for model {}'.format(model_file))


def load_detector(model_file, force_cpu=False):
    """Load a TF or PT detector, depending on the extension of model_file."""
    
    start_time = time.time()
    if model_file.endswith('.pb'):
        from tf_detector import TFDetector
        if force_cpu:
            raise ValueError('force_cpu option is not currently supported for TF detectors, ' + \
                             'use CUDA_VISIBLE_DEVICES=-1')
        detector = TFDetector(model_file)
    elif model_file.endswith('.pt'):
        from pytorch_detector import PTDetector
        detector = PTDetector(model_file, force_cpu, USE_MODEL_NATIVE_CLASSES)        
    else:
        raise ValueError('Unrecognized model format: {}'.format(model_file))
    elapsed = time.time() - start_time
    print('Loaded model in {}'.format(humanfriendly.format_timespan(elapsed)))
    return detector


#%% Main function

def load_and_run_detector(model_file, image_file_names, output_dir,
                          render_confidence_threshold=DEFAULT_RENDERING_CONFIDENCE_THRESHOLD,
                          crop_images=False, box_thickness=DEFAULT_BOX_THICKNESS, 
                          box_expansion=DEFAULT_BOX_EXPANSION, image_size=None
                          ):
    """Load and run detector on target images, and visualize the results."""
    
    if len(image_file_names) == 0:
        print('Warning: no files available')
        return

    print('GPU available: {}'.format(is_gpu_available(model_file)))
    
    detector = load_detector(model_file)

    detection_results = []
    time_load = []
    time_infer = []

    # Dictionary mapping output file names to a collision-avoidance count.
    #
    # Since we'll be writing a bunch of files to the same folder, we rename
    # as necessary to avoid collisions.
    output_filename_collision_counts = {}

    def input_file_to_detection_file(fn, crop_index=-1):
        """Creates unique file names for output files.

        This function does 3 things:
        1) If the --crop flag is used, then each input image may produce several output
            crops. For example, if foo.jpg has 3 detections, then this function should
            get called 3 times, with crop_index taking on 0, 1, then 2. Each time, this
            function appends crop_index to the filename, resulting in
                foo_crop00_detections.jpg
                foo_crop01_detections.jpg
                foo_crop02_detections.jpg

        2) If the --recursive flag is used, then the same file (base)name may appear
            multiple times. However, we output into a single flat folder. To avoid
            filename collisions, we prepend an integer prefix to duplicate filenames:
                foo_crop00_detections.jpg
                0000_foo_crop00_detections.jpg
                0001_foo_crop00_detections.jpg

        3) Prepends the output directory:
                out_dir/foo_crop00_detections.jpg

        Args:
            fn: str, filename
            crop_index: int, crop number

        Returns: output file path
        """
        fn = os.path.basename(fn).lower()
        name, ext = os.path.splitext(fn)
        if crop_index >= 0:
            name += '_crop{:0>2d}'.format(crop_index)
        fn = '{}{}{}'.format(name, image_utils.ImagePathUtils.DETECTION_FILENAME_INSERT, '.jpg')
        if fn in output_filename_collision_counts:
            n_collisions = output_filename_collision_counts[fn]
            fn = '{:0>4d}'.format(n_collisions) + '_' + fn
            output_filename_collision_counts[fn] += 1
        else:
            output_filename_collision_counts[fn] = 0
        fn = os.path.join(output_dir, fn)
        return fn

    # ...def input_file_to_detection_file()
    
    for im_file in tqdm(image_file_names):

        try:
            start_time = time.time()

            image = image_utils.load_image(im_file)

            elapsed = time.time() - start_time
            time_load.append(elapsed)

        except Exception as e:
            print('Image {} cannot be loaded. Exception: {}'.format(im_file, e))
            result = {
                'file': im_file,
                'failure': FAILURE_IMAGE_OPEN
            }
            detection_results.append(result)
            continue

        try:
            start_time = time.time()

            result = detector.generate_detections_one_image(image, im_file,
                       detection_threshold=DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD,
                       image_size=image_size)
            detection_results.append(result)

            elapsed = time.time() - start_time
            time_infer.append(elapsed)

        except Exception as e:
            print('An error occurred while running the detector on image {}. Exception: {}'.format(im_file, e))
            continue

        try:
            if crop_images:

                images_cropped = image_utils.crop_image(result['detections'], image)

                for i_crop, cropped_image in enumerate(images_cropped):
                    output_full_path = input_file_to_detection_file(im_file, i_crop)
                    cropped_image.save(output_full_path)

            else:

                # Image is modified in place
                image_utils.render_detection_bounding_boxes(result['detections'], image,
                            label_map=DEFAULT_DETECTOR_LABEL_MAP,
                            confidence_threshold=render_confidence_threshold,
                            thickness=box_thickness, expansion=box_expansion)
                output_full_path = input_file_to_detection_file(im_file)
                image.save(output_full_path)

        except Exception as e:
            print('Visualizing results on the image {} failed. Exception: {}'.format(im_file, e))
            continue

    # ...for each image

    ave_time_load = statistics.mean(time_load)
    ave_time_infer = statistics.mean(time_infer)
    if len(time_load) > 1 and len(time_infer) > 1:
        std_dev_time_load = humanfriendly.format_timespan(statistics.stdev(time_load))
        std_dev_time_infer = humanfriendly.format_timespan(statistics.stdev(time_infer))
    else:
        std_dev_time_load = 'not available'
        std_dev_time_infer = 'not available'
    print('On average, for each image,')
    print('- loading took {}, std dev is {}'.format(humanfriendly.format_timespan(ave_time_load),
                                                    std_dev_time_load))
    print('- inference took {}, std dev is {}'.format(humanfriendly.format_timespan(ave_time_infer),
                                                      std_dev_time_infer))


import pandas as pd

def getAnimals(manifest):
    if not isinstance(manifest, pd.DataFrame):
        raise AssertionError("'manifest' must be DataFrame.")
    return manifest[manifest['category'].astype(int) == 1].reset_index(drop=True)


def getEmpty(manifest):
    if not isinstance(manifest, pd.DataFrame):
        raise AssertionError("'manifest' must be DataFrame.")
        
    # Removes all images that MegaDetector gave no detection for
    otherdf = manifest[manifest['category'].astype(int) != 1]
    otherdf = otherdf.reset_index(drop=True)
    otherdf['prediction'] = otherdf['category']
    
    # Numbers the class of the non-animals correctly
    if not otherdf.empty:
        otherdf['prediction'] = otherdf['prediction'].replace(2,"human")
        otherdf['prediction'] = otherdf['prediction'].replace(3,"vehicle")
        otherdf['prediction'] = otherdf['prediction'].replace(0,"empty")
        otherdf['confidence'] = otherdf['max_detection_conf']
        
    else:
        otherdf = pd.DataFrame(columns = manifest.columns.values)
                                       
    return otherdf


from tensorflow.keras.models import load_model
from math import ceil
from os.path import isfile
from pandas import DataFrame
import imageCropGenerator

def predictSpecies(detections, model, resize = 456, standardize = False, batch = 1, workers = 1):
  
    if not isfile(model):
        raise AssertionError("The given model file does not exist.")

    model = load_model(model)
        
    if isinstance(detections, DataFrame):
        steps = ceil(len(detections) / batch)
        print(steps)

        filecol = "Frame" if "Frame" in detections.columns else "file"

        if any(detections.columns.isin(["bbox1"])):

            dataset = imageCropGenerator.GenerateCropsFromFile(detections, filecol = filecol,
                                                               resize = resize,
                                                               standardize = standardize, batch = batch)
        else:
            raise AssertionError("Input must be a data frame of crops or vector of file names.")
  
    else:
        raise AssertionError("Input must be a data frame of crops or vector of file names.")
  
    return model.predict(dataset, workers = workers, verbose = 1)
     



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May 17 15:09:03 2023

@author: kyra
"""

import argparse
import sys
import json
import os


imagedir = "/home/kyra/animl-py/examples/Southwest"


    # load the checkpoint if available
    # relative file names are only output at the end; all file paths in the checkpoint are still full paths
    if args.resume_from_checkpoint:
        assert os.path.exists(args.resume_from_checkpoint), 'File at resume_from_checkpoint specified does not exist'
        with open(args.resume_from_checkpoint) as f:
            saved = json.load(f)
        assert 'images' in saved, \
            'The file saved as checkpoint does not have the correct fields; cannot be restored'
        results = saved['images']
        print('Restored {} entries from the checkpoint'.format(len(results)))
    else:
        results = []
    images = imagesFromVideos(args.image_dir, out_dir=args.output_dir,
                              fps=args.fps, frames=args.frames,
                              parallel=args.parallel)

    # test that we can write to the output_file's dir if checkpointing requested
    if args.checkpoint_frequency != -1:
        checkpoint_path = os.path.join(args.output_dir,
                                       'checkpoint_{}.json'.format(datetime.utcnow().strftime("%Y%m%d%H%M%S")))
        with open(checkpoint_path, 'w') as f:
            json.dump({'images': []}, f)
        print('The checkpoint file will be written to {}'.format(checkpoint_path))
    else:
        checkpoint_path = None

    # run MegaDetector and format results
    detections = detectObjectBatch(images, MegaDetector_file, checkpoint_path, 
                                   args.threshold, args.checkpoint_frequency,
                                   results)

    df = parseMD(detections)
    # filter out all non animal detections
    animalDataframe = getAnimals(df)
    otherDataframe = getEmpty(df)

    # Create generator for classification model
    generator = GenerateCropsFromFile(animalDataframe)

    # Create and predict model
    model = keras.models.load_model(Classification_file)
    predictions = model.predict(generator)

    # Format Classification results
    maxDataframe = applyPredictions(animalDataframe, otherDataframe, predictions, classes)
    # Symlink
    symlinkSpecies(maxDataframe, args.output_dir, classes)


if __name__ == '__main__':
    main()



import pandas as pd
import os


def symlinkSpecies(data, linkdir, classes):
    table = pd.read_table(classes, sep=" ", index_col=0)
    for i in range(0, len(table.index)):
        directory = str(table['x'].values[i])
        if not os.path.isdir(linkdir + directory):
            os.makedirs(linkdir + directory)

    for i in range(0, len(data.index)):
        try:
            os.symlink(data.at[i, 'file'],
                       linkdir + data.at[i, 'class'] + "/" + os.path.basename(data.at[i, 'file']))
        except Exception as e:
            print('File already exists. Exception: {}'.format(e))
            continue
            
            
#def symlinkMD():

#der symUnlink():



# -*- coding: utf-8 -*-
"""
Rename and resize images, create a subject set and upload new subjects to it
Created on Mon Jul 11 15:34:53 2016

@author: Mathias Tobler

Copyright (C) 2017 Mathias Tobler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Based on code by:
________________________________________________
Stijn Calders
Space Physics - Space Weather

Royal Belgian Institute for Space Aeronomy (BIRA-IASB)
Ringlaan 3
B-1180 Brussels
BELGIUM

phone  : +32 (0)2 373.04.19
e-mail : stijn.calders@aeronomie.be
web    : www.aeronomie.be
________________________________________________
"""

from panoptes_client import SubjectSet, Subject, Project, Panoptes
import os
import time
import sys
import pandas as pd
from datetime import datetime
from PIL import Image
from PIL import ImageFile

ImageFile.LOAD_TRUNCATED_IMAGES = True


def copy_image(infile, outpath, imageheight):
    img = Image.open(infile.FilePath)
    try:
        exif = img.info['exif']
    except KeyError:
        return None

    wpercent = (imageheight / float(img.size[1]))
    imagewidth = int((float(img.size[0]) * float(wpercent)))
    img = img.resize((imagewidth, imageheight), Image.ANTIALIAS)
    outfile = outpath + infile.FileName
    img.save(outfile, "JPEG", exif=exif)
    return outfile


def create_SubjectSet(project, subject_set_name):
    subject_set = SubjectSet()
    subject_set.links.project = project
    subject_set.display_name = subject_set_name
    subject_set.save()
    return subject_set


def connect_to_Panoptes(usr, pw):
    Panoptes.connect(username=usr, password=pw)
    print("Connected to Zooniverse")


def upload_to_Zooniverse_Simple(project_name, subject_set_name, images, 
                                outdir, imageheight=700):
    sys.stdout.flush()
    # Create a new subject set or append the subjects to an existing one
    project = Project.find(project_name)
    print("Connected to ", project)
    subject_set = SubjectSet.find(int(subject_set_name))
    print(subject_set)
    print("Images to upload: ", images)
    print("Uploading images...")
    for _, infile in images.iterrows():
        if not os.path.exists(infile.FileName):
            print(infile.FileName)
        subject = Subject()  # create new subject
        subject.links.project = project
        outfile = copy_image(infile, outdir, imageheight)
        if outfile is None:
            continue
        subject.add_location(outfile)
        subject.metadata['OrigName'] = infile.FileName
        subject.metadata['!OrigFolder'] = infile.FilePath
        subject.metadata['DateTime'] = str(infile.FileModifyDate)
        subject.metadata['#machine_prediction'] = infile.ZooniverseCode  # map to zooniverse name
        subject.metadata['#machine_confidence'] = infile.confidence
        subject.save()
        subject_set.add(subject)


def upload_to_Zooniverse(project_name, subject_set_name, images, outdir, 
                         maxSeq=1, maxTime=0, imageheight=700):
    sys.stdout.flush()
    # Create a new subject set or append the subjects to an existing one
    project = Project.find(project_name)
    print("Connected to ", project)
    subject_set = SubjectSet.find(int(subject_set_name))

    print(subject_set)
    print("Uploading images...")

    # print(images.columns.values)
    images['DateTime'] = pd.to_datetime(images['DateTime'], format="%Y:%m:%d %H:%M:%S")

    si = 0  # set sequence counter to 0
    for _, infile in images.iterrows():
        bProc = False
        try:
            currentDateTime = time.mktime(time.strptime(infile.DateTime, "%Y-%m-%d %H:%M:%S"))
        except TypeError:
            currentDateTime = datetime.strptime(str(infile.DateTime), "%Y-%m-%d %H:%M:%S")

        # add image to current sequence
        if si > 0 and si < maxSeq and (currentDateTime - dDateTime) <= maxTime and subject.metadata[
                '#machine_prediction'] == infile.ZooniverseCode:
            # print("Continue")
            outfile = copy_image(infile, outdir, imageheight)
            if outfile is None:
                continue
            subject.add_location(outfile)
            subject.metadata['OrigName'] = subject.metadata['OrigName'] + ";" + infile.FileName
            subject.metadata['!OrigFolder'] = subject.metadata['!OrigFolder'] + ";" + infile.FilePath
            subject.metadata['DateTime'] = subject.metadata['DateTime'] + ";" + infile.DateTime.strftime(
                "%Y:%m:%d %H:%M:%S")
            # subject.metadata['NewName'] = subject.metadata['NewName']+";"+ infile.NewName
            dDateTime = currentDateTime
            si = si + 1
            bProc = True
        #  print(subject.metadata)

        # sequence has ended, current image is in a new sequence
        elif 0 < si < maxSeq and ((currentDateTime - dDateTime) > maxTime or subject.metadata[
                '#machine_prediction'] != infile.ZooniverseCode):
            si = maxSeq
            bProc = False

            # need to start a new sequence OR maxSeq = 1
        if si == maxSeq:
            # print("Upload")
            subject.save()
            subject_set.add(subject)
            si = 0

        # first image within a sequence
        if si == 0 and bProc is False:
            subject = Subject()  # create new subject
            subject.links.project = project
            dDateTime = currentDateTime
            outfile = copy_image(infile, outdir, imageheight)
            if outfile is None:
                continue
            subject.add_location(outfile)
            # You can set whatever metadata you want, or none at all
            subject.metadata['OrigName'] = infile.FileName
            subject.metadata['!OrigFolder'] = infile.FilePath
            subject.metadata['DateTime'] = infile.DateTime.strftime("%Y:%m:%d %H:%M:%S")
            subject.metadata['#machine_prediction'] = infile.ZooniverseCode  # map to zooniverse name
            subject.metadata['#machine_confidence'] = infile.confidence
            # subject.metadata['NewName'] = infile.NewName
            si = 1



import pandas as pd
import os
from glob import glob
from datetime import datetime
from imageUtils import is_image


# Extract exif Data and Create File Manifest
#
# @param imagedir file path
# @param exif returns date and time info from exif data, defaults to true
# @param offset add offset to videos, defaults to 0
# @param outfile file path to which the data frame should be saved
#
# @return files dataframe with or without file modify dates
def build_file_manifest(image_dir, exif=True, offset=0, out_file=None):
    if check_file(out_file):
        return load_data(out_file)  # load_data(outfile) load file manifest
    if not os.path.isdir(image_dir):
        raise FileNotFoundError("The given directory does not exist.")

    files = glob(os.path.join(image_dir, '**', '*.*'), recursive=True)

    files = pd.DataFrame(files, columns=["FilePath"])
    files["FileName"] = files["FilePath"].apply(
        lambda x: os.path.split(x)[1])
    files["FileModifyDate"] = files["FilePath"].apply(
        lambda x: datetime.fromtimestamp(
            os.path.getmtime(x)).strftime('%Y-%m-%d %H:%M:%S'))

    if(out_file):
        save_data(files, out_file)

    return files


# Set Working Directory and Save File Global Variables
#
# @param workingdir local directory that contains data to process
# @param pkg.env environment to create global variables in
#
# @return None
class WorkingDirectory():
    def __init__(self, working_dir):
        if not os.path.isdir(working_dir):
            raise FileNotFoundError("The given directory does not exist.")
        if not working_dir.endswith("/"):
            working_dir = working_dir + "/"

        self.basedir = working_dir + "Animl-Directory/"
        self.datadir = self.basedir + "Data/"
        self.cropdir = self.basedir + "Crops/"
        self.vidfdir = self.basedir + "Frames/"
        self.linkdir = self.basedir + "Sorted/"

        # Create directories if they do not already exist
        if not os.path.exists(self.datadir):
            os.makedirs(self.datadir)
        if not os.path.exists(self.cropdir):
            os.makedirs(self.cropdir)
        if not os.path.exists(self.vidfdir):
            os.makedirs(self.vidfdir)
        if not os.path.exists(self.linkdir):
            os.makedirs(self.linkdir)

        # Assign specific file paths
        self.filemanifest = self.datadir + "FileManifest.csv"
        self.imageframes = self.datadir + "ImageFrames.csv"
        self.results = self.datadir + "Results.csv"
        self.crops = self.datadir + "Crops.csv"
        self.predictions = self.datadir + "Predictions.csv"
        self.mdresults = self.datadir + "mdresults.csv"


# Save Data to Given File
#
# @param data the dataframe to be saved
# @param file the full path of the saved file
# @param prompt if true, prompts the user to confirm overwrite
#
# @return none
def save_data(data, file, prompt=True):
    if os.path.exists(file) and (prompt is True):
        prompt = "Output file exists, would you like to overwrite? y/n: "
        if input(prompt).lower() == "y":
            data.to_csv(file, index=False)
    else:
        data.to_csv(file, index=False)


# Load .csv File
#
# @param file the full path of the file to load
#
# @return data extracted from the file
def load_data(file):
    ext = os.path.splitext(file)[1]
    if ext == ".csv":
        return pd.read_csv(file)
    else:
        raise AssertionError("Error. Expecting a .csv file.")


# Check for files existence and prompt user if they want to load
#
# @param file the full path of the file to check
#
# @return a boolean indicating wether a file was found
#             and the user wants to load or not
def check_file(file):
    if (file is not None) and os.path.isfile(file):
        date = os.path.getmtime(file)
        date = datetime.fromtimestamp(date)
        prompt = "Output file already exists and was last modified {}, \
                 would you like to load it? y/n: ".format(date)
        if input(prompt).lower() == "y":
            return(True)
    return(False)
